""" Module used to blend ini, environment, and explicit dictionary properties
to determine application configuration. Some hard coded defaults for Galaxy but
this should be reusable by tool shed and pulsar as well.
"""
import os
import os.path
import sys
from configparser import (
    BasicInterpolation,
    ConfigParser,
    InterpolationError,
)
from functools import partial
from itertools import (
    product,
    starmap,
)
from typing import (
    cast,
    Iterable,
    Optional,
)

import yaml

from galaxy.exceptions import InvalidFileFormatError
from galaxy.util.path import (
    extensions,
    has_ext,
    joinext,
)


def get_from_env(key: str, prefixes: Iterable[str], default: Optional[str] = None):
    """
    Return first available value for prefix+key set in the environment, or default.
    An empty prefix is ignored.

    Useful when we need to check against multiple prefixes sequentially,
    returning the first available value.
    """
    for prefix in prefixes:
        if prefix:
            value = os.getenv(f"{prefix}{key}")
            if value:
                return value
    return default


def find_config_file(names, exts=None, dirs=None, include_samples=False):
    """Locate a config file in multiple directories, with multiple extensions.

    >>> from shutil import rmtree
    >>> from tempfile import mkdtemp
    >>> def touch(d, f):
    ...     open(os.path.join(d, f), 'w').close()
    >>> def _find_config_file(*args, **kwargs):
    ...     return find_config_file(*args, **kwargs).replace(d, '')
    >>> d = mkdtemp()
    >>> d1 = os.path.join(d, 'd1')
    >>> d2 = os.path.join(d, 'd2')
    >>> os.makedirs(d1)
    >>> os.makedirs(d2)
    >>> touch(d1, 'foo.ini')
    >>> touch(d1, 'foo.bar')
    >>> touch(d1, 'baz.ini.sample')
    >>> touch(d2, 'foo.yaml')
    >>> touch(d2, 'baz.yml')
    >>> _find_config_file('foo', dirs=(d1, d2))
    '/d1/foo.ini'
    >>> _find_config_file('baz', dirs=(d1, d2))
    '/d2/baz.yml'
    >>> _find_config_file('baz', dirs=(d1, d2), include_samples=True)
    '/d2/baz.yml'
    >>> _find_config_file('baz', dirs=(d1,), include_samples=True)
    '/d1/baz.ini.sample'
    >>> _find_config_file('foo', dirs=(d2, d1))
    '/d2/foo.yaml'
    >>> find_config_file('quux', dirs=(d,))
    >>> _find_config_file('foo', exts=('bar', 'ini'), dirs=(d1,))
    '/d1/foo.bar'
    >>> rmtree(d)
    """
    found = __find_config_files(
        names,
        exts=exts or extensions["yaml"] + extensions["ini"],
        dirs=dirs or [os.getcwd(), os.path.join(os.getcwd(), "config")],
        include_samples=include_samples,
    )
    if not found:
        return None
    # doesn't really make sense to log here but we should probably generate a warning of some kind if more than one
    # config is found.
    return found[0]


def load_app_properties(
    kwds=None, ini_file=None, ini_section=None, config_file=None, config_section=None, config_prefix="GALAXY_CONFIG_"
):
    if config_file is None:
        config_file = ini_file
        config_section = ini_section

    # read from file or init w/no file
    if config_file:
        properties = read_properties_from_file(config_file, config_section)
    else:
        properties = {"__file__": None}

    # update from kwds
    if kwds:
        properties.update(kwds)

    # update from env
    override_prefix = f"{config_prefix}OVERRIDE_"
    for key in os.environ:
        if key.startswith(override_prefix):
            config_key = key[len(override_prefix) :].lower()
            properties[config_key] = os.environ[key]
        elif key.startswith(config_prefix):
            config_key = key[len(config_prefix) :].lower()
            if config_key not in properties:
                properties[config_key] = os.environ[key]

    return properties


def read_properties_from_file(config_file, config_section=None):
    properties = {}
    if has_ext(config_file, "yaml", aliases=True, ignore="sample"):
        if config_section is None:
            config_section = "galaxy"
        properties.update(__default_properties(config_file))
        raw_properties = _read_from_yaml_file(config_file)
        if raw_properties:
            properties.update(raw_properties.get(config_section) or {})
    elif has_ext(config_file, "ini", aliases=True, ignore="sample"):
        if config_section is None:
            config_section = "app:main"
        parser = nice_config_parser(config_file)  # default properties loaded w/parser
        if parser.has_section(config_section):
            properties.update(dict(parser.items(config_section)))
        else:
            properties.update(parser.defaults())
    else:
        raise InvalidFileFormatError(f"File '{config_file}' doesn't have a supported extension")
    return properties


def _read_from_yaml_file(path):
    with open(path) as f:
        return yaml.safe_load(f)


def nice_config_parser(path):
    parser = NicerConfigParser(path, defaults=__default_properties(path))
    with open(path) as f:
        parser.read_file(f)
    return parser


class _InterpolateWrapper(BasicInterpolation):
    def before_get(self, parser, section, option, value, defaults):
        try:
            return super().before_get(parser, section, option, value, defaults)
        except InterpolationError:
            e = cast(InterpolationError, sys.exc_info()[1])
            args = list(e.args)
            args[0] = f"Error in file {parser.filename}: {e}"
            e.args = tuple(args)
            e.message = args[0]
            raise


class NicerConfigParser(ConfigParser):
    def __init__(self, filename, *args, **kw):
        kw["interpolation"] = _InterpolateWrapper()
        ConfigParser.__init__(self, *args, **kw)
        self.filename = filename

    def optionxform(self, optionstr: str) -> str:
        # Don't lower-case keys
        return str(super().optionxform(optionstr))

    def defaults(self):
        """Return the defaults, with their values interpolated (with the
        defaults dict itself)

        Mainly to support defaults using values such as %(here)s
        """
        defaults = dict(ConfigParser.defaults(self))
        for key, val in defaults.items():
            defaults[key] = self.get("DEFAULT", key) or val
        return defaults


def _running_from_source():
    paths = ["run.sh", "lib/galaxy/__init__.py", "scripts/common_startup.sh"]
    return all(map(os.path.exists, paths))


running_from_source = _running_from_source()


def get_data_dir(properties):
    data_dir = properties.get("data_dir", None)
    if data_dir is None:
        if running_from_source:
            data_dir = "./database"
        else:
            config_dir = properties.get("config_dir", os.path.dirname(properties["__file__"]))
            data_dir = os.path.join(config_dir, "data")
    return data_dir


def __get_all_configs(dirs, names):
    return list(filter(os.path.exists, starmap(os.path.join, product(dirs, names))))


def __find_config_files(names, exts=None, dirs=None, include_samples=False):
    sample_names: Iterable[str] = []
    if isinstance(names, str):
        names = [names]
    if not dirs:
        dirs = [os.getcwd()]
    if exts:
        # add exts to names, converts back into a list because it's going to be small and we might consume names twice
        names = list(starmap(joinext, product(names, exts)))
    if include_samples:
        sample_names = map(partial(joinext, ext="sample"), names)
    # check for all names in each dir before moving to the next dir. could do it the other way around but that makes
    # less sense to me.
    return __get_all_configs(dirs, names) or __get_all_configs(dirs, sample_names)


def __default_properties(path):
    return {"here": os.path.dirname(os.path.abspath(path)), "__file__": os.path.abspath(path)}


__all__ = ("find_config_file", "get_data_dir", "load_app_properties", "NicerConfigParser", "running_from_source")



"""This module defines Galaxy's custom exceptions.

A Galaxy exception is an exception that extends :class:`MessageException` which
defines an HTTP status code (represented by the `status_code` attribute) and a
default error message.

New exceptions should be defined by adding an entry to `error_codes.json` in this
directory to define a default error message and a Galaxy "error code". A concrete
Python class should be added in this file defining an HTTP status code (as
`status_code`) and error code (`error_code`) object loaded dynamically from
`error_codes.json`.

Reflecting Galaxy's origins as a web application, these exceptions tend to be a
bit web-oriented. However this module is a dependency of modules and tools that
have nothing to do with the web - keep this in mind when defining exception names
and messages.
"""

from .error_codes import (
    error_codes_by_name,
    ErrorCode,
)


class MessageException(Exception):
    """Most generic Galaxy exception - indicates merely that some exceptional condition happened."""

    # status code to be set when used with API.
    status_code: int = 400
    # Error code information embedded into API json responses.
    err_code: ErrorCode = error_codes_by_name["UNKNOWN"]

    def __init__(self, err_msg=None, type="info", **extra_error_info):
        self.err_msg = err_msg or self.err_code.default_error_message
        self.type = type
        self.extra_error_info = extra_error_info

    @staticmethod
    def from_code(status_code, message):
        exception_class = MessageException
        if status_code == 404:
            exception_class = ObjectNotFound
        elif status_code / 100 == 5:
            exception_class = InternalServerError
        return exception_class(message)

    def __str__(self):
        return self.err_msg


class ItemDeletionException(MessageException):
    pass


class ObjectInvalid(Exception):
    """Accessed object store ID is invalid"""


# Please keep the exceptions ordered by status code


class AcceptedRetryLater(MessageException):
    status_code = 202
    err_code = error_codes_by_name["ACCEPTED_RETRY_LATER"]
    retry_after: int

    def __init__(self, msg, retry_after=60):
        super().__init__(msg)
        self.retry_after = retry_after


class NoContentException(MessageException):
    status_code = 204
    err_code = error_codes_by_name["NO_CONTENT_GENERIC"]


class ActionInputError(MessageException):
    status_code = 400
    err_code = error_codes_by_name["USER_REQUEST_INVALID_PARAMETER"]

    def __init__(self, err_msg, type="error"):
        super().__init__(err_msg, type)


class DuplicatedSlugException(MessageException):
    status_code = 400
    err_code = error_codes_by_name["USER_SLUG_DUPLICATE"]


class DuplicatedIdentifierException(MessageException):
    status_code = 400
    err_code = error_codes_by_name["USER_IDENTIFIER_DUPLICATE"]


class ObjectAttributeInvalidException(MessageException):
    status_code = 400
    err_code = error_codes_by_name["USER_OBJECT_ATTRIBUTE_INVALID"]


class ObjectAttributeMissingException(MessageException):
    status_code = 400
    err_code = error_codes_by_name["USER_OBJECT_ATTRIBUTE_MISSING"]


class MalformedId(MessageException):
    status_code = 400
    err_code = error_codes_by_name["MALFORMED_ID"]


class UserInvalidRunAsException(MessageException):
    status_code = 400
    err_code = error_codes_by_name["USER_INVALID_RUN_AS"]


class MalformedContents(MessageException):
    status_code = 400
    err_code = error_codes_by_name["MALFORMED_CONTENTS"]


class UnknownContentsType(MessageException):
    status_code = 400
    err_code = error_codes_by_name["UNKNOWN_CONTENTS_TYPE"]


class RequestParameterMissingException(MessageException):
    status_code = 400
    err_code = error_codes_by_name["USER_REQUEST_MISSING_PARAMETER"]


class ToolMetaParameterException(MessageException):
    status_code = 400
    err_code = error_codes_by_name["USER_TOOL_META_PARAMETER_PROBLEM"]


class ToolMissingException(MessageException):
    status_code = 400
    err_code = error_codes_by_name["USER_TOOL_MISSING_PROBLEM"]

    def __init__(self, err_msg=None, type="info", tool_id=None, **extra_error_info):
        super().__init__(err_msg, type, **extra_error_info)
        self.tool_id = tool_id


class RequestParameterInvalidException(MessageException):
    status_code = 400
    err_code = error_codes_by_name["USER_REQUEST_INVALID_PARAMETER"]


class ToolInputsNotReadyException(MessageException):
    status_code = 400
    error_code = error_codes_by_name["TOOL_INPUTS_NOT_READY"]


class RealUserRequiredException(MessageException):
    status_code = 400
    error_code = error_codes_by_name["REAL_USER_REQUIRED"]


class AuthenticationFailed(MessageException):
    status_code = 401
    err_code = error_codes_by_name["USER_AUTHENTICATION_FAILED"]


class AuthenticationRequired(MessageException):
    status_code = 403
    # TODO: as 401 and send WWW-Authenticate: ???
    err_code = error_codes_by_name["USER_NO_API_KEY"]


class ItemAccessibilityException(MessageException):
    status_code = 403
    err_code = error_codes_by_name["USER_CANNOT_ACCESS_ITEM"]


class ItemOwnershipException(MessageException):
    status_code = 403
    err_code = error_codes_by_name["USER_DOES_NOT_OWN_ITEM"]


class ItemImmutableException(MessageException):
    status_code = 403
    err_code = error_codes_by_name["ITEM_IS_IMMUTABLE"]


class ConfigDoesNotAllowException(MessageException):
    status_code = 403
    err_code = error_codes_by_name["CONFIG_DOES_NOT_ALLOW"]


class InsufficientPermissionsException(MessageException):
    status_code = 403
    err_code = error_codes_by_name["INSUFFICIENT_PERMISSIONS"]


class UserCannotRunAsException(MessageException):
    status_code = 403
    err_code = error_codes_by_name["USER_CANNOT_RUN_AS"]


class AdminRequiredException(MessageException):
    status_code = 403
    err_code = error_codes_by_name["ADMIN_REQUIRED"]


class UserActivationRequiredException(MessageException):
    status_code = 403
    err_code = error_codes_by_name["USER_ACTIVATION_REQUIRED"]


class ObjectNotFound(MessageException):
    """Accessed object was not found"""

    status_code = 404
    err_code = error_codes_by_name["USER_OBJECT_NOT_FOUND"]


class Conflict(MessageException):
    status_code = 409
    err_code = error_codes_by_name["CONFLICT"]


class DeprecatedMethod(MessageException):
    """
    Method (or a particular form/arg signature) has been removed and won't be available later
    """

    status_code = 410
    err_code = error_codes_by_name["DEPRECATED_API_CALL"]


class ConfigurationError(Exception):
    status_code = 500
    err_code = error_codes_by_name["CONFIG_ERROR"]


class InconsistentDatabase(MessageException):
    status_code = 500
    err_code = error_codes_by_name["INCONSISTENT_DATABASE"]


class InternalServerError(MessageException):
    status_code = 500
    err_code = error_codes_by_name["INTERNAL_SERVER_ERROR"]


class ToolExecutionError(MessageException):
    status_code = 500
    err_code = error_codes_by_name["TOOL_EXECUTION_ERROR"]

    def __init__(self, err_msg, type="error", job=None):
        super().__init__(err_msg, type)
        self.job = job


class NotImplemented(MessageException):
    status_code = 501
    err_code = error_codes_by_name["NOT_IMPLEMENTED"]


class InvalidFileFormatError(MessageException):
    status_code = 500
    err_code = error_codes_by_name["INVALID_FILE_FORMAT"]


class ReferenceDataError(MessageException):
    status_code = 500
    err_code = error_codes_by_name["REFERENCE_DATA_ERROR"]


class ServerNotConfiguredForRequest(MessageException):
    # A bit like ConfigDoesNotAllowException but it has nothing to do with the user of the
    # request being "forbidden". It just isn't configured.
    status_code = 501
    err_code = error_codes_by_name["SERVER_NOT_CONFIGURED_FOR_REQUEST"]


class HandlerAssignmentError(Exception):
    def __init__(self, msg=None, obj=None, **kwargs):
        super().__init__(msg)
        self.obj = obj



"""
Utility functions for bi-directional Python version compatibility.  Python 2.5
introduced hashlib which replaced sha in Python 2.4 and previous versions.
"""

import hashlib
import hmac
import logging
import os
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from . import smart_str

log = logging.getLogger(__name__)

BLOCK_SIZE = 1024 * 1024

HashFunctionT = Callable[[], "hashlib._Hash"]

sha1 = hashlib.sha1
sha256 = hashlib.sha256
sha512 = hashlib.sha512
sha = sha1
md5 = hashlib.md5


class HashFunctionNameEnum(str, Enum):
    """Particular pieces of information that can be requested for a dataset."""

    md5 = "MD5"
    sha1 = "SHA-1"
    sha256 = "SHA-256"
    sha512 = "SHA-512"


HASH_NAME_ALIAS: Dict[str, str] = {
    "SHA1": "SHA-1",
    "SHA256": "SHA-256",
    "SHA512": "SHA-512",
}

HASH_NAME_MAP: Dict[HashFunctionNameEnum, HashFunctionT] = {
    HashFunctionNameEnum.md5: md5,
    HashFunctionNameEnum.sha1: sha1,
    HashFunctionNameEnum.sha256: sha256,
    HashFunctionNameEnum.sha512: sha512,
}
HASH_NAMES: List[HashFunctionNameEnum] = list(HASH_NAME_MAP.keys())


def memory_bound_hexdigest(
    hash_func: Optional[HashFunctionT] = None,
    hash_func_name: Optional[HashFunctionNameEnum] = None,
    path: Optional[str] = None,
    file=None,
):
    if hash_func is None:
        assert hash_func_name is not None
        hash_func = HASH_NAME_MAP[hash_func_name]

    hasher = hash_func()
    if file is None:
        assert path is not None
        file = open(path, "rb")
    else:
        assert path is None, "Cannot specify path and path keyword arguments."

    try:
        for block in iter(lambda: file.read(BLOCK_SIZE), b""):
            hasher.update(block)
        return hasher.hexdigest()
    finally:
        file.close()


def md5_hash_file(path: Union[str, os.PathLike]) -> Optional[str]:
    """
    Return a md5 hashdigest for a file or None if path could not be read.
    """
    hasher = hashlib.md5()
    try:
        with open(path, "rb") as afile:
            buf = afile.read()
            hasher.update(buf)
            return hasher.hexdigest()
    except OSError:
        # This may happen if path has been deleted
        return None


def md5_hash_str(s):
    """
    Return hex encoded md5 hash of string s
    """
    m = hashlib.md5()
    m.update(smart_str(s))
    return m.hexdigest()


def new_secure_hash_v2(text_type: Union[bytes, str]) -> str:
    """More modern version of new_secure_hash.

    Certain passwords are set via new_insecure_hash (previously new_secure_hash),
    so that needs to remain for legacy purposes.
    """
    assert text_type is not None
    return sha512(smart_str(text_type)).hexdigest()


def new_insecure_hash(text_type: Union[bytes, str]) -> str:
    """Returns the hexdigest of the sha1 hash of the argument `text_type`.

    Previously called new_secure_hash, but this should not be considered
    secure - SHA1 is no longer considered a secure hash and has been broken
    since the early 2000s.

    use_pbkdf2 should be set by default and galaxy.security.passwords should
    be the default used for passwords in Galaxy.
    """
    assert text_type is not None
    return sha1(smart_str(text_type)).hexdigest()


def hmac_new(key: Union[bytes, str], value: Union[bytes, str]) -> str:
    return hmac.new(smart_str(key), smart_str(value), sha).hexdigest()


def is_hashable(value: Any) -> bool:
    try:
        hash(value)
    except Exception:
        return False
    return True


def parse_checksum_hash(checksum: str) -> Tuple[HashFunctionNameEnum, str]:
    """Parses checksum strings in the form of `hash_type$hash_value` considering possible aliases."""
    hash_name, hash_value = checksum.split("$", 1)
    hash_name = hash_name.upper()
    if hash_name in HASH_NAME_ALIAS:
        hash_name = HASH_NAME_ALIAS[hash_name]
    if hash_name not in HASH_NAMES:
        raise ValueError(f"Unsupported hash function '{hash_name}'. Supported functions: [{','.join(HASH_NAMES)}]")
    return HashFunctionNameEnum(hash_name), hash_value


__all__ = (
    "md5",
    "hashlib",
    "sha1",
    "sha",
    "new_insecure_hash",
    "new_secure_hash_v2",
    "hmac_new",
    "is_hashable",
    "parse_checksum_hash",
)



"""
Classes for generating HTML forms
"""

import logging

from galaxy.util import asbool

log = logging.getLogger(__name__)


class BaseField:
    def __init__(self, name, value=None, label=None, **kwds):
        self.name = name
        self.label = label
        self.value = value
        self.disabled = kwds.get("disabled", False)
        if "optional" in kwds:
            self.optional = asbool(kwds.get("optional"))
        else:
            self.optional = kwds.get("required", "optional") == "optional"
        self.help = kwds.get("helptext")

    def to_dict(self):
        return {
            "name": self.name,
            "label": self.label,
            "disabled": self.disabled,
            "optional": self.optional,
            "value": self.value,
            "help": self.help,
        }


class TextField(BaseField):
    """
    A standard text input box.
    """

    def to_dict(self):
        d = super().to_dict()
        d["type"] = "text"
        return d


class PasswordField(BaseField):
    """
    A password input box. text appears as "******"
    """

    def to_dict(self):
        d = super().to_dict()
        d["type"] = "password"
        return d


class TextArea(BaseField):
    """
    A standard text area box.
    """

    def to_dict(self):
        d = super().to_dict()
        d["type"] = "text"
        d["area"] = True
        return d


class CheckboxField(BaseField):
    """
    A checkbox (boolean input)
    """

    @staticmethod
    def is_checked(value):
        if value in [True, "True", "true"]:
            return True
        return False

    def to_dict(self):
        d = super().to_dict()
        d["type"] = "boolean"
        return d


class SelectField(BaseField):
    """
    A select field.
    """

    def __init__(
        self,
        name,
        multiple=None,
        display=None,
        field_id=None,
        value=None,
        selectlist=None,
        refresh_on_change=False,
        **kwds,
    ):
        super().__init__(name, value, **kwds)
        self.field_id = field_id
        self.multiple = multiple or False
        self.refresh_on_change = refresh_on_change
        self.selectlist = selectlist or []
        self.options = list()
        if display == "checkboxes":
            assert multiple, "Checkbox display only supported for multiple select"
        elif display == "radio":
            assert not (multiple), "Radio display only supported for single select"
        elif display is not None:
            raise Exception(f"Unknown display type: {display}")
        self.display = display

    def add_option(self, label, value, selected=False):
        self.options.append((str(label), value, selected))

    def to_dict(self):
        d = super().to_dict()
        d["type"] = "select"
        d["display"] = self.display
        d["multiple"] = self.multiple
        d["data"] = []
        for value in self.selectlist:
            d["data"].append({"label": value, "value": value})
        d["options"] = [{"label": t[0], "value": t[1]} for t in self.options]
        return d


class AddressField(BaseField):
    @staticmethod
    def fields():
        return [
            ("desc", "Short address description", "Required"),
            ("name", "Name", ""),
            ("institution", "Institution", ""),
            ("address", "Address", ""),
            ("city", "City", ""),
            ("state", "State/Province/Region", ""),
            ("postal_code", "Postal Code", ""),
            ("country", "Country", ""),
            ("phone", "Phone", ""),
        ]

    def __init__(self, name, user=None, value=None, security=None, **kwds):
        super().__init__(name, value, **kwds)
        self.user = user
        self.security = security

    def to_dict(self):
        d = super().to_dict()
        d["type"] = "select"
        d["data"] = []
        if self.user and self.security:
            for a in self.user.addresses:
                if not a.deleted:
                    d["data"].append({"label": a.desc, "value": self.security.encode_id(a.id)})
        return d


class WorkflowField(BaseField):
    def __init__(self, name, user=None, value=None, security=None, **kwds):
        super().__init__(name, value, **kwds)
        self.user = user
        self.value = value
        self.security = security

    def to_dict(self):
        d = super().to_dict()
        d["type"] = "select"
        d["data"] = []
        if self.user and self.security:
            for a in self.user.stored_workflows:
                if not a.deleted:
                    d["data"].append({"label": a.name, "value": self.security.encode_id(a.id)})
        return d


class WorkflowMappingField(BaseField):
    def __init__(self, name, user=None, value=None, **kwds):
        super().__init__(name, value, **kwds)
        self.user = user


class HistoryField(BaseField):
    def __init__(self, name, user=None, value=None, security=None, **kwds):
        super().__init__(name, value, **kwds)
        self.user = user
        self.value = value
        self.security = security

    def to_dict(self):
        d = super().to_dict()
        d["type"] = "select"
        d["data"] = [{"label": "New History", "value": "new"}]
        if self.user and self.security:
            for a in self.user.histories:
                if not a.deleted:
                    d["data"].append({"label": a.name, "value": self.security.encode_id(a.id)})
        return d
